
// This software is written by HKJ from lygte-info.dk

// It is a skeleton for handling 16 bit pwm and depending on processor there will be 2, 3 or 11 available
// Analog A0 is used as a test input.

#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega168P__) || defined(__AVR_ATmega328__) || defined(__AVR_ATmega328P__)
// This is for UNO and Nano processor
const byte pwm0pin = 9;     // PB1, OC1A
const byte pwm1pin = 10;    // PB2, OC1B
#elif defined(__AVR_ATmega32U4__)|| defined(__AVR_ATmega16U4__)
// This is for Leonardo and ProMicro processors
const byte pwm0pin = 9;     // PB5, OC1A
const byte pwm1pin = 10;    // PB6, OC1B
const byte pwm2pin = 5;     // PC6, OC3A
#elif defined(__AVR_ATmega2560__)
const byte pwm0pin = 11;     // PB5, OC1A
const byte pwm1pin = 12;     // PB6, OC1B
const byte pwm2pin = 5;      // PE5, OC3A
const byte pwm3pin = 2;      // PE4, OC3B
const byte pwm4pin = 3;      // PE5, OC3C
const byte pwm5pin = 6;      // PH3, OC4A
const byte pwm6pin = 7;      // PH4, OC4B
const byte pwm7pin = 8;      // PH5, OC4C
const byte pwm8pin = 46;     // PL3, OC5A
const byte pwm9pin = 45;    // PL4, OC5B
const byte pwm10pin = 44;    // PL5, OC5C
#endif

// This is the maximum PWM value and means the PWM frequency is around 244Hz.
// Reducing this value will increase the PWM frequency and reduce the number of steps, this also means the pow() function must be recalculated.
const uint16_t MAX_PWM_VALUE = 65535U;

//----------------------------------------------------------------------------------------------------
// Setup timers for 16 bit PWM
void initPWM() {
  noInterrupts();
  TCCR1A = 1 << WGM11 | 1 << COM1A1 | 1 << COM1B1; // set on top, clear OC on compare match
  TCCR1B = 1 << CS10  | 1 << WGM12 | 1 << WGM13;   // clk/1, mode 14 fast PWM
  ICR1 = MAX_PWM_VALUE;
#if defined(__AVR_ATmega32U4__)|| defined(__AVR_ATmega16U4__)
  TCCR3A =  1 << COM3A1 | 1 << WGM31; // Clear OC on compare match, set on top
  TCCR3B = 1 <<  CS30 | 1 << WGM32 | 1 << WGM33;  // clk/1, mode 14 fast PWM
  ICR3 = MAX_PWM_VALUE;
#elif defined(__AVR_ATmega2560__)
  TCCR3A = 1 << WGM31 | 1 << COM3A1 | 1 << COM3B1 | 1 << COM3C1; // set on top, clear OC on compare match
  TCCR3B = 1 << CS30  | 1 << WGM32 | 1 << WGM33;   // clk/1, mode 14 fast PWM
  TCCR4A = 1 << WGM41 | 1 << COM4A1 | 1 << COM4B1 | 1 << COM4C1; // set on top, clear OC on compare match
  TCCR4B = 1 << CS40  | 1 << WGM42 | 1 << WGM43;   // clk/1, mode 14 fast PWM
  TCCR5A = 1 << WGM51 | 1 << COM5A1 | 1 << COM5B1 | 1 << COM5C1; // set on top, clear OC on compare match
  TCCR5B = 1 << CS50  | 1 << WGM52 | 1 << WGM53;   // clk/1, mode 14 fast PWM
  ICR3 = MAX_PWM_VALUE;
  ICR4 = MAX_PWM_VALUE;
  ICR5 = MAX_PWM_VALUE;
#endif
  interrupts();
  pinMode(pwm0pin, OUTPUT);
  pinMode(pwm1pin, OUTPUT);
#if defined(__AVR_ATmega32U4__)|| defined(__AVR_ATmega16U4__)
  pinMode(pwm2pin, OUTPUT);
#elif defined(__AVR_ATmega2560__)
  pinMode(pwm2pin, OUTPUT);
  pinMode(pwm3pin, OUTPUT);
  pinMode(pwm4pin, OUTPUT);
  pinMode(pwm5pin, OUTPUT);
  pinMode(pwm6pin, OUTPUT);
  pinMode(pwm7pin, OUTPUT);
  pinMode(pwm8pin, OUTPUT);
  pinMode(pwm9pin, OUTPUT);
  pinMode(pwm10pin, OUTPUT);
#endif
}

//----------------------------------------------------------------------------------------------------
// Set a 16 bit PWM value for a channel
void setPWM(byte no, uint16_t pwm) {
  noInterrupts();
  switch (no) {
    case 0 :
      OCR1A = pwm;
      break;
    case 1 :
      OCR1B = pwm;
      break;
#if defined(__AVR_ATmega32U4__)|| defined(__AVR_ATmega16U4__)
    case 2 :
      OCR3A = pwm;
      break;
#elif defined(__AVR_ATmega2560__)      
    case 2 :
      OCR3A = pwm;
      break;
    case 3 :
      OCR3B = pwm;
      break;
    case 4 :
      OCR3C = pwm;
      break;
    case 5 :
      OCR4A = pwm;
      break;
    case 6 :
      OCR4B = pwm;
      break;
    case 7 :
      OCR4C = pwm;
      break;
    case 8 :
      OCR5A = pwm;
      break;
    case 9 :
      OCR5B = pwm;
      break;
    case 10 :
      OCR5C = pwm;
      break;
#endif
  }
  interrupts();
}

//----------------------------------------------------------------------------------------------------
// All these functions use a logarithmic scale, this means they are ideal for brightness adjustments:
// I have defined 3 ranges 0..100, 0..255 and 0..1023, use the findMaxForPow() function to make others
void setPWM100(byte no, double value) {
  setPWM(no, pow(value, 2.4082365036));
};

void setPWM255(byte no, double value) {
  setPWM(no, pow(value, 2.0014095306));
};

void setPWM1023(byte no, double value) {

  setPWM(no, pow(value, 1.6002234458));
};

//----------------------------------------------------------------------------------------------------
// Generate a constant for pow() for a specified maximum input value
void findMaxForPow(uint16_t maxInputValue) {
  double exp = 1;
  double expLast = 0;
  double expadj = 1;
  double v = 0;

  while (exp != expLast) {
    expLast = exp;
    v = pow(maxInputValue, exp);
    if (v > MAX_PWM_VALUE) {
      exp -= expadj;
      expadj = expadj / 10;
    }
    exp += expadj;
  }
  Serial.print(F("For a maximum input value of: "));
  Serial.print(maxInputValue);
  Serial.print(F("  Use: "));
  Serial.print(exp, 10);
  Serial.println(F("  In pow(..) function "));
}

//----------------------------------------------------------------------------------------------------
void setup() {
  Serial.begin(9600);
  while (!Serial);
  Serial.println("Starting...");
  initPWM();
  //findMaxForPow(255);
}

//----------------------------------------------------------------------------------------------------
// Use A0 as test input
void loop() {
  //  setPWM(0,map(analogRead(A0),0,1023,0,MAX_PWM_VALUE));
  setPWM1023(10, analogRead(A0));
  delay(50);

}
